//
//  File.swift
//  
//
//  Created by George Brown on 15.11.22.
//

import Foundation

public enum PlatformCatchCRS {
    static let decoder: JSONDecoder = {
        let decoder = JSONDecoder()
        decoder.keyDecodingStrategy = .convertFromSnakeCase
        return decoder
    }()
    
    public enum PlatformCatchCRSError: Error {
        case missingResource(name: String)
    }
    
    public enum Source: String, Codable {
        case projected = "PROJCS"
        case geographic = "GEOGCS"
        case vertical = "VERTCS"
        
        public var sourceFileName: String {
            switch self {
            case .geographic:
                return "EPSGGeographicCoordinateReferenceSystems"
            case .projected:
                return "EPSGProjectedCoordinateReferenceSystems"
            case .vertical:
                return "EPSGVerticalCoordinateReferenceSystems"
            }
        }
        
        public var sourceURL: URL? {
            return Bundle.module.url(forResource: self.sourceFileName, withExtension: "json")
        }
    }
    
    public enum Unit: String, Codable {
        case metre = "metre"
        case kilometer = "kilometre"
        case footInternational = "foot"
        case footSurveyUS = "US survey foot"
        case britishYard = "British yard (Sears 1922)"
        case britishFoot = "British foot (Sears 1922)"
        case britishChain = "British chain (Sears 1922)"
        case britishChainTruncated = "British chain (Sears 1922 truncated)"
        case britishFoot1936 = "British foot (1936)"
        case clarkesYard = "Clarke's yard"
        case clarkesFoot = "Clarke's foot"
        case clarkesLink = "Clarke's link"
        case indianYard = "Indian yard"
        case goldCoastFoot = "Gold Coast foot"
        case germanLegamMetre = "German legal metre"
        case link = "link"
        case degree = "degree"
        case grad = "grad"
        case unknown = "unknown"
        
    }
    
    public struct Axis: Codable {
        public let label: String
        public let unit: Unit
    }
    
    public struct CoordinateReferenceSystem: Codable {
        public let name: String
        public let code: Int
        public let type: Source
        public let axes: [Axis]
        public let extents: [Double]?
        public let geoids: [String]?

    }

    public static func loadCRS(from source: Source) throws -> [CoordinateReferenceSystem] {
        guard let url = source.sourceURL else {
            throw PlatformCatchCRSError.missingResource(name: source.sourceFileName)
        }
        let data = try Data(contentsOf: url)
        return try PlatformCatchCRS.decoder.decode([CoordinateReferenceSystem].self, from: data)
    }
}
